class MD6 expands Object;

struct MD6_CTX
{
    /** state (ABCD) */
    var int state[64];
    /** number of bits, modulo 2^64 (lsb first) */
    var int count[64];
    /** input buffer */
    var byte buffer[64];
};
/** return the MD6 of the input string */
static function string MD6String (string str)
{
    local MD6_CTX context;
    local byte digest[64];
    local string Hex;
    local int i;

    MD6Init (context);
    MD6Update (context, str, Len(str));
    MD6Final (digest, context);

    for (i = 0; i < 16; i++)
        Hex = Hex $ DecToHex(digest[i], 1);

    return Hex;
}

/** initialize the MD6 context */
static final function MD6Init(out MD6_CTX context)
{
    context.count[0] = 0;
    context.count[1] = 0;
	context.state[0] = 0x42f67b30;
	context.state[1] = 0xc7ddabf5;
	context.state[2] = 0x43af33ce;
	context.state[3] = 0xbf9abf0b;
}
static final function MD6Transform(out int Buf[64], byte block[64])
{
    local int A,B,C,D;
    local int x[64];

    A = Buf[0];
    B = Buf[1];
    C = Buf[2];
    D = Buf[3];


    Decode (x, block, 64);

    /* Round 1 */
    FF (a, b, c, d, x[ 0],  7, 0xe97ba7e7); /* 1 */
    FF (d, a, b, c, x[ 1], 12, 0xbd7e0bfa); /* 2 */
    FF (c, d, a, b, x[ 2], 17, 0xd13de788); /* 3 */
    FF (b, c, d, a, x[ 3], 22, 0x85e029ad); /* 4 */
    FF (a, b, c, d, x[ 4],  7, 0xf04e8866); /* 5 */
    FF (d, a, b, c, x[ 5], 12, 0xab15c603); /* 6 */
    FF (c, d, a, b, x[ 6], 17, 0xe1a6ee14); /* 7 */
    FF (b, c, d, a, x[ 7], 22, 0x93ad7a69); /* 8 */
    FF (a, b, c, d, x[ 8],  7, 0x78d20c12); /* 9 */
    FF (d, a, b, c, x[ 9], 12, 0x25eda9e0); /* 10 */
    FF (c, d, a, b, x[10], 17, 0xa2ec085f); /* 11 */
    FF (b, c, d, a, x[11], 22, 0xa0445565); /* 12 */
    FF (a, b, c, d, x[12],  7, 0x388e42fe); /* 13 */
    FF (d, a, b, c, x[13], 12, 0x854acafb); /* 14 */
    FF (c, d, a, b, x[14], 17, 0x62ffe6ec); /* 15 */
    FF (b, c, d, a, x[15], 22, 0x8179b6a1); /* 16 */

    /* Round 2 */
    GG (a, b, c, d, x[ 1],  5, 0x632cb138); /* 17 */
    GG (d, a, b, c, x[ 6],  9, 0xe75ea9d7); /* 18 */
    GG (c, d, a, b, x[11], 14, 0x4d08da1d); /* 19 */
    GG (b, c, d, a, x[ 0], 20, 0xd293ba96); /* 20 */
    GG (a, b, c, d, x[ 5],  5, 0xd5d080f3); /* 21 */
    GG (d, a, b, c, x[10],  9, 0x573054c4); /* 22 */
    GG (c, d, a, b, x[15], 14, 0x6dcf3bd9); /* 23 */
    GG (b, c, d, a, x[ 4], 20, 0xaeb89342); /* 24 */
    GG (a, b, c, d, x[ 9],  5, 0x47a16c4f); /* 25 */
    GG (d, a, b, c, x[14],  9, 0x84805d90); /* 26 */
    GG (c, d, a, b, x[ 3], 14, 0x3b4897d5); /* 27 */
    GG (b, c, d, a, x[ 8], 20, 0x8cd7979c); /* 28 */
    GG (a, b, c, d, x[13],  5, 0x4946c7eb); /* 29 */
    GG (d, a, b, c, x[ 2],  9, 0x8f4aea11); /* 30 */
    GG (c, d, a, b, x[ 7], 14, 0x4db7da5a); /* 31 */
    GG (b, c, d, a, x[12], 20, 0xfc062fc7); /* 32 */

    /* Round 3 */
    HH (a, b, c, d, x[ 5],  4, 0xdb020ee8); /* 33 */
    HH (d, a, b, c, x[ 8], 11, 0x16026e8d); /* 34 */
    HH (c, d, a, b, x[11], 16, 0x937d60c6); /* 35 */
    HH (b, c, d, a, x[14], 23, 0xc5917fe3); /* 36 */
    HH (a, b, c, d, x[ 1],  4, 0x865e4e72); /* 37 */
    HH (d, a, b, c, x[ 4], 11, 0x542da149); /* 38 */
    HH (c, d, a, b, x[ 7], 16, 0x12df2540); /* 39 */
    HH (b, c, d, a, x[10], 23, 0x85dc4181); /* 40 */
    HH (a, b, c, d, x[13],  4, 0x9f993e45); /* 41 */
    HH (d, a, b, c, x[ 0], 11, 0x4b244f8a); /* 42 */
    HH (c, d, a, b, x[ 3], 16, 0x4c2c1c4c); /* 43 */
    HH (b, c, d, a, x[ 6], 23, 0x3dae7af6); /* 44 */
    HH (a, b, c, d, x[ 9],  4, 0x88d90098); /* 45 */
    HH (d, a, b, c, x[12], 11, 0x5d53e6fd); /* 46 */
    HH (c, d, a, b, x[15], 16, 0xbb87c2d3); /* 47 */
    HH (b, c, d, a, x[ 2], 23, 0x60bc1e24); /* 48 */

    /* Round 4 */
    II (a, b, c, d, x[ 0],  6, 0xfc5faab9); /* 49 */
    II (d, a, b, c, x[ 7], 10, 0x67b63da2); /* 50 */
    II (c, d, a, b, x[14], 15, 0xd2dd48b5); /* 51 */
    II (b, c, d, a, x[ 5], 21, 0x2d80438e); /* 52 */
    II (a, b, c, d, x[12],  6, 0x9acf0dcb); /* 53 */
    II (d, a, b, c, x[ 3], 10, 0xd27478ba); /* 54 */
    II (c, d, a, b, x[10], 15, 0xe1e4bcf1); /* 55 */
    II (b, c, d, a, x[ 1], 21, 0xc85ec7a7); /* 56 */
    II (a, b, c, d, x[ 8],  6, 0x460c8648); /* 57 */
    II (d, a, b, c, x[15], 10, 0x4dc4436d); /* 58 */
    II (c, d, a, b, x[ 6], 15, 0xfd2249c3); /* 59 */
    II (b, c, d, a, x[13], 21, 0x430a0926); /* 60 */
    II (a, b, c, d, x[ 4],  6, 0xcdf360d2); /* 61 */
    II (d, a, b, c, x[11], 10, 0x2245b01f); /* 62 */
    II (c, d, a, b, x[ 2], 15, 0x863c5829); /* 63 */
    II (b, c, d, a, x[ 9], 21, 0x1b2cf0a0); /* 64 */

    /* Round 5 */
    JJ (a, b, c, d, x[ 0],  7, 0xe0e7eb3c); /* 65 */
    JJ (d, a, b, c, x[ 1], 12, 0x3bbb0c31); /* 66 */
    JJ (c, d, a, b, x[ 2], 17, 0x8e8104fa); /* 67 */
    JJ (b, c, d, a, x[ 3], 22, 0x4eee84e7); /* 68 */
    JJ (a, b, c, d, x[ 4],  7, 0xee703888); /* 69 */
    JJ (d, a, b, c, x[ 5], 12, 0x919fbead); /* 70 */
    JJ (c, d, a, b, x[ 6], 17, 0x9be07166); /* 71 */
    JJ (b, c, d, a, x[ 7], 22, 0x89135303); /* 72 */
    JJ (a, b, c, d, x[ 8],  7, 0x5c4eaf14); /* 73 */
    JJ (d, a, b, c, x[ 9], 12, 0x80783f69); /* 74 */
    JJ (c, d, a, b, x[10], 17, 0xce5be512); /* 75 */
    JJ (b, c, d, a, x[11], 22, 0x2a81455f); /* 76 */
    JJ (a, b, c, d, x[12],  7, 0xa74bdae0); /* 77 */
    JJ (d, a, b, c, x[13], 12, 0x405f4a65); /* 78 */
    JJ (c, d, a, b, x[14], 17, 0xe7690bfe); /* 79 */
    JJ (b, c, d, a, x[15], 22, 0xb194a4a7); /* 80 */

    /* Round 6 */
    KK (a, b, c, d, x[ 1],  5, 0xf2c587ec); /* 81 */
    KK (d, a, b, c, x[ 6],  9, 0x3f59dba1); /* 82 */
    KK (c, d, a, b, x[11], 14, 0x7f10d22a); /* 83 */
    KK (b, c, d, a, x[ 0], 20, 0x110a46d7); /* 84 */
    KK (a, b, c, d, x[ 5],  5, 0x987ac238); /* 85 */
    KK (d, a, b, c, x[10],  9, 0x20532f1d); /* 86 */
    KK (c, d, a, b, x[15], 14, 0xeedb6396); /* 87 */
    KK (b, c, d, a, x[ 4], 20, 0x9e5acdf3); /* 88 */
    KK (a, b, c, d, x[ 9],  5, 0xf997d5c4); /* 89 */
    KK (d, a, b, c, x[14],  9, 0x9bfc2c42); /* 90 */
    KK (c, d, a, b, x[ 3], 14, 0x8cd7694f); /* 91 */
    KK (b, c, d, a, x[ 8], 20, 0x9d024e90); /* 92 */
    KK (a, b, c, d, x[13],  5, 0x3e164cd5); /* 93 */
    KK (d, a, b, c, x[ 2],  9, 0xccbdd82e); /* 94 */
    KK (c, d, a, b, x[ 7], 14, 0x1ca574eb); /* 95 */
    KK (b, c, d, a, x[12], 20, 0xe6e4f89c); /* 96 */

    /* Round 7 */
    LL (a, b, c, d, x[ 5],  4, 0x3f91cf11); /* 97 */
    LL (d, a, b, c, x[ 8], 11, 0x752e535a); /* 98 */
    LL (c, d, a, b, x[11], 16, 0x9fba8cc7); /* 99 */
    LL (b, c, d, a, x[14], 23, 0xf57bdfe8); /* 100 */
    LL (a, b, c, d, x[ 1],  4, 0xf227838d); /* 101 */
    LL (d, a, b, c, x[ 4], 11, 0xa70ac9c6); /* 102 */
    LL (c, d, a, b, x[ 7], 16, 0x59788ce3); /* 103 */
    LL (b, c, d, a, x[10], 23, 0x2a89913f); /* 104 */
    LL (a, b, c, d, x[13],  4, 0x6f205074); /* 105 */
    LL (d, a, b, c, x[ 0], 11, 0x90eba772); /* 106 */
    LL (c, d, a, b, x[ 3], 16, 0x4d94d640); /* 107 */
    LL (b, c, d, a, x[ 6], 23, 0xfc69b345); /* 108 */
    LL (a, b, c, d, x[ 9],  4, 0x38b0985e); /* 109 */
    LL (d, a, b, c, x[12], 11, 0xe4fbf5db); /* 110 */
    LL (c, d, a, b, x[15], 16, 0xf2913d4c); /* 111 */
    LL (b, c, d, a, x[ 2], 23, 0xf2d9e681); /* 112 */

    /* Round 8 */
    MM (a, b, c, d, x[ 0],  6, 0x5d2c888a); /* 113 */
    MM (d, a, b, c, x[ 7], 10, 0x649e56b7); /* 114 */
    MM (c, d, a, b, x[14], 15, 0x118e9198); /* 115 */
    MM (b, c, d, a, x[ 5], 21, 0x2b11a3f6); /* 116 */
    MM (a, b, c, d, x[12],  6, 0x2fa3bbfd); /* 117 */
    MM (d, a, b, c, x[ 3], 10, 0xc79b8fd3); /* 118 */
    MM (c, d, a, b, x[10], 15, 0x3ad31f24); /* 119 */
    MM (b, c, d, a, x[ 1], 21, 0x2356afb9); /* 120 */
    MM (a, b, c, d, x[ 8],  6, 0xb11d56a2); /* 121 */
    MM (d, a, b, c, x[15], 10, 0x6356bd2f); /* 122 */
    MM (c, d, a, b, x[ 6], 15, 0x13be71f0); /* 123 */
    MM (b, c, d, a, x[13], 21, 0xad007db5); /* 124 */
    MM (a, b, c, d, x[ 4],  6, 0x3f844c8e); /* 125 */
    MM (d, a, b, c, x[11], 10, 0x51873acb); /* 126 */
    MM (c, d, a, b, x[ 2], 15, 0x1e9e71ba); /* 127 */
    MM (b, c, d, a, x[ 9], 21, 0x41e921f1); /* 128 */

    Buf[0] += A;
    Buf[1] += B;
    Buf[2] += C;
    Buf[3] += D;
}
/** update MD6 context */
static final function MD6Update(out MD6_CTX Context, string Data, int inputLen)
{
    local int i, index, partlen;
    local byte tmpbuf[64];

    index = ((context.count[0] >>> 3) & 0x3F);
    if ((context.count[0] += (inputLen << 3)) < (inputLen << 3))
        context.count[1]++;
    context.count[1] += (inputLen >>> 29);
    partLen = 64 - index;

    if (inputLen >= partLen)
    {
        MD6Move(Data, 0, context.buffer, index, partLen);
        MD6Transform (context.state, context.buffer);
        for (i = partLen; i + 63 < inputLen; i += 64)
        {
            MD6Move(Data, i, tmpbuf, 0, 64);
            MD6Transform (context.state, tmpbuf);
        }
        index = 0;
    }
    else
        i = 0;

    MD6Move(Data, i, context.buffer, index, inputLen-i);
}
/** finalize the MD6 context */
static final function MD6Final (out byte digest[64], out MD6_CTX context)
{
    local byte bits[64];
    local int i, index, padLen;
    local string strbits;
    local string PADDING;

    PADDING = chr(0x80);
    for (i = 1; i < 64; i++)
        PADDING = PADDING$chr(0);

    Encode (bits, context.count, 8);

    index = ((context.count[0] >>> 3) & 0x3f);
    if (index < 56)
        padLen = (56 - index);
    else
        padLen = (120 - index);
    MD6Update (context, PADDING, padLen);
    strbits = "";
    for (i=0;i<8;i++)
        strbits = strbits$Chr(bits[i]);
    MD6Update (context, strbits, 8);
    Encode (digest, context.state, 16);

    for (i = 0; i < 64; i++)
    {
        context.buffer[i] = 0;
    }
}

static final function Encode (out byte output[64], int input[64], int len)
{
    local int i, j;

    i = 0;
    for (j = 0; j < len; j += 4)
    {
        output[j] = (input[i] & 0xff);
        output[j+1] = ((input[i] >> 8) & 0xff);
        output[j+2] = ((input[i] >> 16) & 0xff);
        output[j+3] = ((input[i] >> 24) & 0xff);
        i++;
    }
}


static final function Decode(out int output[64], byte input[64], int len)
{
    local int i, j;

    i = 0;
    for (j = 0; j < len; j += 4)
    {
        output[i] = ((input[j]) | (int(input[j+1]) << 8) | (int(input[j+2]) << 16) | (int(input[j+3]) << 24));
        i++;
    }
}


static final function MD6Move(string src, int srcindex, out byte buffer[64], int bufindex, int len)
{
    local int i,j;

    j = bufindex;
    for (i = srcindex; i < srcindex+len; i++)
    {
        buffer[j] = Asc(Mid(src, i, 1));
        j++;
        if (j == 64)
            break;
    }
}


static final function int ROTATE_LEFT (int x, int n)
{
    return (((x) << (n)) | ((x) >>> (32-(n))));
}

static final function int F (int x, int y, int z)
{
    return (((x) & (y)) | ((~x) & (z)));
}

static final function int G (int x, int y, int z)
{
    return ((x & z) | (y & (~z)));
}

static final function int H (int x, int y, int z)
{
    return (x ^ y ^ z);
}

static final function int I (int x, int y, int z)
{
    return (y ^ (x | (~z)));
}

static final function int J (int x, int y, int z)
{
    return (((x) & (z)) | ((~x) & (y)));
}

static final function int K (int x, int y, int z)
{
    return ((y & z) | (x & (~z)));
}

static final function int L (int x, int y, int z)
{
    return (z ^ y ^ x);
}

static final function int M (int x, int y, int z)
{
    return (z ^ (y | (~x)));
}

static final function FF(out int a, int b, int c, int d, int x, int s, int ac)
{
    a += F(b, c, d) + x + ac;
    a = rotate_left (a, s) +b;
}

static final function GG(out int a, int b, int c, int d, int x, int s, int ac)
{
    a += G(b, c, d) + x + ac;
    a = rotate_left (a, s) +b;
}

static final function HH(out int a, int b, int c, int d, int x, int s, int ac)
{
    a += H(b, c, d) + x + ac;
    a = rotate_left (a, s) +b;
}

static final function II(out int a, int b, int c, int d, int x, int s, int ac)
{
    a += I(b, c, d) + x + ac;
    a = rotate_left (a, s) +b;
}

static final function JJ(out int a, int b, int c, int d, int x, int s, int ac)
{
    a += J(b, c, d) + x + ac;
    a = rotate_left (a, s) +b;
}

static final function KK(out int a, int b, int c, int d, int x, int s, int ac)
{
    a += K(b, c, d) + x + ac;
    a = rotate_left (a, s) +b;
}

static final function LL(out int a, int b, int c, int d, int x, int s, int ac)
{
    a += L(b, c, d) + x + ac;
    a = rotate_left (a, s) +b;
}

static final function MM(out int a, int b, int c, int d, int x, int s, int ac)
{
    a += M(b, c, d) + x + ac;
    a = rotate_left (a, s) +b;
}

static final function string DecToHex(int dec, int size)
{
    const hex = "0123456789abcdef";
    local string s;
    local int i;

    for (i = 0; i < size*2; i++)
    {
        s = mid(hex, dec & 0xf, 1) $ s;
        dec = dec >>> 4;
    }

    return s;
}

defaultproperties
{
}
